// SPDX-License-Identifier: MIT

pragma solidity ^0.8.0;

contract DepositDAO {
    string private _name;
    uint256 private _minMembers;
    uint256 private _maxMembers;
    uint256 private _depositAmount;
    address private _depositLeader;
    
    enum DepositStatus { Opened, Closed, Canceled }
    DepositStatus private _dStatus;
    
    Voters internal voters;
    struct Voters {
        mapping (address => VoterInfo) voterInfo;
        address[] keys;
    }
    struct VoterInfo {
        uint256 deposit;
        bool isRegistered;
        bool hasVote;
    }
    
    enum ProposalStatus { Opened, Passed, NonPassed, Canceled }
    enum Opinion { Yes, No, None }
    
    struct TransferProposal {
        uint idx;
    	address proposer;
    	address receiver;
        uint256 value;
    	ProposalStatus pStatus;
    	string description;
        uint yesCount;
        uint noCount;
        uint totalVotes;
    }
    TransferProposal[] private tProposals;
    mapping(uint256 => mapping(address => Opinion)) tOpinions;
    
    constructor (string memory name_, uint256 minMembers_, uint256 maxMembers_, uint256 depositAmount_) payable {
        require(msg.value == depositAmount_);
        address sender = msg.sender;
        _name = name_;
        _minMembers = minMembers_;
        _maxMembers = maxMembers_;
        _depositAmount = depositAmount_;
        _depositLeader = msg.sender;
        _dStatus = DepositStatus.Opened;
        
        voters.voterInfo[sender].deposit = msg.value;
        voters.voterInfo[sender].isRegistered = true;
        voters.voterInfo[sender].hasVote = true;
        voters.keys.push(sender);
    }
    
    event AddMember(address indexed member, uint256 deposit);
    event TransferETH(address indexed to, uint256 value);
    event ChangeDepositStatus(DepositStatus from, DepositStatus to);
    
    
    modifier onlyLeader() {
        require(msg.sender == _depositLeader);
        _;
    }
    
    modifier onlyOpen() {
        require(_dStatus == DepositStatus.Opened);
        _;
    }
    
    modifier onlyMember() {
        require(voters.voterInfo[msg.sender].isRegistered);
        _;
    }

    modifier onlyHasVote() {
        require(voters.voterInfo[msg.sender].hasVote);
        _;
    }

    
    function joinDeposit() onlyOpen payable public returns (bool) {
        require(!voters.voterInfo[msg.sender].isRegistered, 
            "the voter is already registered");
        require(voters.keys.length < _maxMembers,
            "this depositDAO is full");
        require(msg.value == _depositAmount, "NOT enough ETH");
        
        address sender = msg.sender;
        voters.voterInfo[sender].deposit = msg.value;
        voters.voterInfo[sender].isRegistered = true;
        voters.voterInfo[sender].hasVote = true;
        voters.keys.push(sender);
        
        emit AddMember(sender, _depositAmount);

        return true;
    }
    
    function chageDepositStatus() onlyLeader public {
        DepositStatus prevStatus = _dStatus;
        
        if (_dStatus == DepositStatus.Opened) {
            if (voters.keys.length < _minMembers) {
                // refund all
                if (voters.keys.length > 0) {
                    for (uint i = 0; i < voters.keys.length; i++) {
                        address recipient = voters.keys[i];
                        uint256 deposit = voters.voterInfo[voters.keys[i]].deposit;
                        _transferEther(payable(recipient), deposit);
                    }
                }
                _dStatus = DepositStatus.Canceled;
            } else {
                _dStatus = DepositStatus.Closed;
            }
        } else {
            _dStatus = DepositStatus.Opened;
        }
        emit ChangeDepositStatus(prevStatus, _dStatus);
    }
    
    
    function getMembers() public view returns (address[] memory, uint256[] memory, bool[] memory, bool[] memory) {
        address[] memory addrs = new address[](voters.keys.length);
        uint256[] memory deps = new uint256[](voters.keys.length);
        bool[] memory regs = new bool[](voters.keys.length);
        bool[] memory votes = new bool[](voters.keys.length);
        
        for (uint i = 0; i < voters.keys.length; i++ ){
            addrs[i] = voters.keys[i];
            deps[i] = voters.voterInfo[voters.keys[i]].deposit;
            regs[i] = voters.voterInfo[voters.keys[i]].isRegistered;
            votes[i] = voters.voterInfo[voters.keys[i]].hasVote;
        }
        
        return (addrs, deps, regs, votes);
    }
    
    
    // View Functions of the information of DepositDAO contract
    
    function name() public view returns (string memory) {
        return _name;
    }
    function minMembers() public view returns (uint256) {
        return _minMembers;
    }
    function maxMembers() public view returns (uint256) {
        return _maxMembers;
    }
    function curMembers() public view returns (uint256) {
        return voters.keys.length;
    }
    function depositAmount() public view returns (uint256) {
        return _depositAmount;
    }
    function depositStatus() public view returns (DepositStatus) {
        return _dStatus;
    }
    function balance() public view returns (uint256) {
        return address(this).balance;
    }
    
    // Add Transfer Proposal 
    function addTransferProposal (address _receiver, uint256 _amount, string memory _description) 
    onlyMember public returns(bool) {
        uint pIndex = this.getTransferProposalsNumber();
        uint totalVotes = 0;
        
        for (uint i = 0; i< voters.keys.length; i++) {
            address voterAddr = voters.keys[i];
            
            if (voters.voterInfo[voterAddr].hasVote) {
                tOpinions[tProposals.length][voterAddr] = Opinion.None;
                totalVotes ++;
            }
        }

        tProposals.push(TransferProposal({
            idx: pIndex,
            pStatus: ProposalStatus.Opened,
            description: _description,
            proposer: msg.sender,
            receiver: _receiver,
            value: _amount,
            yesCount: 0,
            noCount: 0,
            totalVotes: totalVotes
        }));

        return true;
    }
    
    
    function getTransferProposalsNumber() public view returns (uint) {
        return tProposals.length;
    }
    
    function getAllTransferProposals() onlyMember public view returns (uint[] memory index, address[] memory proposer, 
    address[] memory receiver, uint256[] memory value, ProposalStatus[] memory status, string[] memory description) {
        uint pLength = this.getTransferProposalsNumber();
        
        uint[] memory _idx = new uint[](pLength);
        address[] memory _proposer = new address[](pLength);
        address[] memory _receiver = new address[](pLength);
        uint256[] memory _value = new uint256[](pLength);
        ProposalStatus[] memory _pStatus = new ProposalStatus[](pLength);
        string[] memory _description = new string[](pLength);
        
        for(uint i=0; i < pLength; i++) {
            _idx[i] = tProposals[i].idx;
            _proposer[i] = tProposals[i].proposer;
            _receiver[i] = tProposals[i].receiver;
            _value[i] = tProposals[i].value;
            _pStatus[i] = tProposals[i].pStatus;
            _description[i] = tProposals[i].description;
        }
        return (_idx, _proposer, _receiver, _value, _pStatus, _description);
    }

    function getTansferProposalInfo(uint idx) onlyMember public view returns 
    (address proposer, ProposalStatus status, string memory description) {
        TransferProposal storage p = tProposals[idx];
        return (p.proposer, p.pStatus, p.description);
    }

    function getTransferProposalResult(uint idx) onlyMember public view returns 
    (uint yesCount, uint noCount, uint totalVotes) {
        TransferProposal storage p = tProposals[idx];
        return (p.yesCount, p.noCount, p.totalVotes);
    }

    function cancelProposal(uint idx ) public returns (bool) {
        
        TransferProposal storage t = tProposals[idx];
        require(t.proposer == msg.sender && t.pStatus == ProposalStatus.Opened);
        t.pStatus = ProposalStatus.Canceled;
        
        return true;
    }

    function closeProposal (uint idx ) onlyLeader public returns(bool) {
        uint voteCount = 0;

        TransferProposal storage t = tProposals[idx];
        require(t.pStatus == ProposalStatus.Opened);
        voteCount = t.yesCount + t.noCount;
        
        if (2*voteCount > t.totalVotes && 2*t.yesCount > voteCount) {
            if (voters.voterInfo[t.proposer].deposit >= t.value ) {
                _transferEther(payable(t.receiver), t.value);
                voters.voterInfo[t.proposer].deposit -= t.value;
                t.pStatus = ProposalStatus.Passed;
            } else {
                t.pStatus = ProposalStatus.Canceled;
            }
        } else {
            t.pStatus = ProposalStatus.NonPassed;
        }

        return true;
    }
    
    function voteToProposal(uint idx, bool opinion) onlyHasVote public returns(bool) {
        
        TransferProposal storage t = tProposals[idx];
        require(t.pStatus == ProposalStatus.Opened,
            "This proposal is not opened");
        require(tOpinions[idx][msg.sender] == Opinion.None,
            "This account already voted");

        if(opinion) {
            tOpinions[idx][msg.sender] = Opinion.Yes;
            t.yesCount ++;
        } else {
            tOpinions[idx][msg.sender] = Opinion.No;
            t.noCount ++;
        }

        return true;
    }
    

    // Transfer Ether from contract to receiver
    function _transferEther(address payable recipient, uint _amount) private {
        require(address(this).balance >= _amount, 
            "The value of Ether in this contract is less than request");
        recipient.transfer(_amount);
        
        emit TransferETH(recipient, _amount);
    }
    
}